{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name deinterlace}
{@description De-interlace image by splitting into two frames}

{$I directives.pas}
uses simage,getparam;
var par        : tparams;
    cc         : CSImg_Config;
    img1,img2  : TSImg_FPImage;
    fi,fo1,fo2 : ansistring;
    t          : double;
    s          : boolean;
    mi         : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='<outfile1> <outfile2>';
        fo1:='';   defparam('<outfile1>','First output image, which is the even half frame',@fo1,nil,tpastr,rapdreq);
        fo2:='';   defparam('<outfile2>','Second output image, which is the odd half frame',@fo2,nil,tpastr,rapdreq);
        fi:='';    defparam('-i','Input image (default: stdin)',fi);
                   defparam('-h','Help',helpf);
        t:=1;      defparam('-t','Threshold; interpolate only, if deviation is larger then threshold; (default: 0.1)',t);
        s:=false;  defparam('-s','Swap half frames',s);
        run;
        done;
    end;

    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('deinterlace: values: source='+fi+' dest1='+fo1+' dest2='+fo2+' t='+real2str(t)+' s='+boolean2str(s));

    mi:=CSImg_MetaInfo.create;
    cc.Info('deinterlace: Reading input image '+fi);
    SImg_ReadImageFile(cc,fi,img1,mi);

    cc.Info('deinterlace: Splitting frame');
    SImg_DeInterlace(cc, img1, img1,img2, t,s);

    cc.Info('deinterlace: Writing 1st output image ' + fo1);
    SImg_WriteImageFile(cc,fo1,img1,mi);
    SImg_Free(cc,img1);

    cc.Info('deinterlace: Writing 2nd output image ' + fo2);
    SImg_WriteImageFile(cc,fo2,img2,mi);
    SImg_Free(cc,img2);

    cc.destroy;
end.
