{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name filter}
{@description Applies a filter function.}

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    fi                  : TSImg_FPImage;
    ifn,ofn             : ansistring;
    rad,quant           : double;
    fg,fq,fa,dd,db      : boolean;
    fra,frr,frq         : boolean;
    i,ss                : longint;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';      defparam('<infile>','Source image (default: stdin)',ifn);
        ofn:='';      defparam('<outfile>','Destination image (default: stdout)',ofn);
                      defparam('-h','Help',helpf);
        rad:=1.5;     defparam('-r','Radius (default: 1.5)', rad);
        ss:=1;        defparam('-/','Down-sample by this factor (ignored by drop filters)', ss);
        fg:=false;    defparam('-g','Use Gauss low-pass filter (default, y=exp(-sqr(r/rad)) )',fg);
        fa:=false;    defparam('-a','Calculate average of a square region',fa);
        quant:=0.5;   defparam('-q','Calculate quantile of a square region',quant);
        fra:=false;   defparam('-ra','Calculate average of a round region',fra);
                      defparam('-rq','Calculate quantile of a round region',quant);
        frr:=false;   defparam('-rr','Calculate reciprocal average of a round region',frr);
        dd:=false;    defparam('-d','Drop darkest value of square region',dd);
        db:=false;    defparam('-b','Drop brightest value of square region',db);
                      defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
        run;
        fq:=paramdef('-q');
        frq:=paramdef('-rq');

        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);

    i:=0;
    if fq then i+=1;
    if fg then i+=1;
    if fa then i+=1;
    if dd or db then i+=1;
    if fra then i+=1;
    if frq then i+=1;
    if frr then i+=1;
    if i>1 then cc.error('More than one filter defined');

    i:=0;
    if frr then i:=-5;
    if frq then i:=-4;
    if fra then i:=-3;
    if fq  then i:=-2;
    if fa  then i:=-1;
    if dd  then i:=1;
    if db  then i:=i or 2;

    cc.debug('filter: values: source='+ifn+' dest='+ofn+' rad='+real2str(rad)+' algo='+int2str(i)+' ss='+int2str(ss)+' quant='+real2str(quant));

    fi.pt:=nil;
    cc.Info('filter: Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,fi);

    cc.Info('filter: Applying filter using ' +int2str(cc.threads) + ' threads');
    SImg_Filter(cc,fi,fi,rad,i,ss,quant);

    cc.Info('filter: Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,fi);

    SImg_Free(cc,fi);
    cc.destroy;
end.

