{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@ame getalignment2}
{@description Aligns a diffuse bright source to center}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    fi,fo,fa            : ansistring;
    img                 : TSImg_FPImage;
    res,res2            : TSImg_AlignmentParams;
    mi                  : CSImg_MetaInfo;
    rad,cx,cy,x2,y2,a   : double;
    cr                  : longint;
    cxd,cyd,ia,ic       : boolean;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<infile> [<outfile>]';
        fi:='';          defparam('<infile>','Reference source list',fi);
        fo:='';          defparam('<outfile>','Output file (header only).',fo);
        rad:=10;         defparam('-r','Approximate radius of diffuse source (default: 10)',rad);
        cx:=0;           defparam('-cx','Reference X position (default: image center)',cx);
        cy:=0;           defparam('-cy','Reference Y position (default: image center)',cy);
        cr:=-1;          defparam('-cr','Search with that radius from reference position (default: 1/2 of the image diagonal)',cr);
        fa:='';          defparam('-a','Alignment file. Output is the corrected alignment file.\n Reference position is the position in the reference image', fa);
        ia:=false;       defparam('-ia','Outputs the absolute position', ia);
        ic:=false;       defparam('-ic','Outputs the position relative to center', ic);
                         defparam('-h','Help',helpf);
                         defparam('-o','Source list containing alignment information',fo); deffmt('-o','<outfile>');
        run;
        cxd:=vardef(@cx);
        cyd:=vardef(@cy);
        done;
    end;

    cc.debug('getalignment2: in='+fi+' out='+fo+'  align='+fa+'  rad='+real2str(rad)+'  cx='+real2str(cx)+'  cy='+real2str(cy)+'  cr='+int2str(cr));
    mi:=CSImg_MetaInfo.create();


    cc.Info('getalignment2: Reading input image '+fi);
    SImg_ReadImageFile(cc,fi,img);

    if not cxd then cx:=img.nx div 2;
    if not cyd then cy:=img.ny div 2;
    if cr<0 then cr:=round(0.5*sqrt(sqr(img.nx)+sqr(img.ny)));

    if fa<>'' then
    begin
        { alignment data }
        cc.Info('sdbadd: Reading alignment data '+fa);
        SImg_RHeader(cc,fa,mi);
        if not SImg_ReadAlignmentParams(cc,res2,mi) then cc.error('getalignment2: parameters not set');

        { alignment: coordinate transformation }
        x2:=sqr(cx);
        y2:=sqr(cy);
        a:= res2.d*(x2+y2) + res2.c1*cx + res2.c2*cy;
        cx+=cx*a + res2.c3*x2 + res2.c5*y2;
        cy+=cy*a + res2.c4*y2 + res2.c6*x2;
        x2:=res2.b11*cx + res2.b12*cy + res2.ax - res2.xmin;
        cy:=res2.b21*cx + res2.b22*cy + res2.ay - res2.ymin;
        cx:=x2;
//      writeln(stderr,'cx: ',cx:0:2,' cy: ',cy:0:2);
    end;

    cc.Info('getalignment2: Calculating alignment using '+int2str(cc.threads) + ' threads');
    SImg_GetAlignment2(cc, img, rad, res, round(cx),round(cy),cr);
    res.weight:=1;
    cc.info('getalignment2: Source position: (' + real2str(-res.xmin,-1) + ', ' + real2str(-res.ymin,-1) + ')');

    if ia then writeln('pos_a: ',-res.xmin:0:1, ' ', -res.ymin:0:1);
    if ic then writeln('pos_c: ',(1-img.nx)*0.5-res.xmin:0:1, ' ', (1-img.ny)*0.5-res.ymin:0:1);

    mi.reset;
    if fa='' then SImg_WriteAlignmentParams(res,mi)
    else begin
        res2.ax-=cx+res.xmin;
        res2.ay-=cy+res.ymin;
        SImg_WriteAlignmentParams(res2,mi);
    end;

    if (fo<>'') or (fa<>'') or ((not ia) and (not ic)) then
    begin
        cc.Info('getalignment2: Writing alignment data '+fo);
        SImg_WFITSHeader(cc,CSImg_FileOutput.init(fo),mi).close(cc);
    end;

    mi.destroy;
    SImg_Free(cc,img);
    cc.destroy;
end.
