{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name getsources}
{@description Find (point) sources of an image}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,wfn,ofn         : ansistring;
    n                   : longint;
    xc,yc               : double;
    mi                  : CSImg_MetaInfo;
    fi,fw               : TSImg_FPImage;
    pl                  : TSImg_RDImage;
    cmp                 : boolean;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';      defparam('<infile>','Input image (default: stdin)',ifn);
        ofn:='';      defparam('<outfile>','Output source list (default: stdout)',ofn);
                      defparam('-h','Help',helpf);
        n:=1000;      defparam('-n','Maximum number of sources (default: 1000)',n);
        xc:=-1;       defparam('-xc','Distortion center in X direction (default: image center, relevant for alignment)',xc);
        yc:=-1;       defparam('-yc','Distortion center in Y direction (default: image center, relevant for alignment)',yc);
        wfn:='';      defparam('-w','Pixel weight image',wfn);  deffmt('-w','<infile>');
        cmp:=false;   defparam('-c','Compressed output without chains',cmp);
                      defparam('-o','Output source list',ofn);    deffmt('-o','<outfile>');
        run;

        cc.debug('getsources: in='+ifn+' out='+ofn+' n='+int2str(n)+' cmp='+boolean2str(cmp));

        { input image }
        fi.pt:=nil;
        mi:=CSImg_MetaInfo.create();
        cc.Info('getsources: Reading input image '+ifn);
        SImg_ReadImageFile(cc,ifn,fi,mi);
        if not paramdef('-xc') then xc:=(fi.nx-1)*0.5;
        if not paramdef('-yc') then yc:=(fi.ny-1)*0.5;

        done;
    end;

    { weight image }
    fw.pt:=nil;
    if wfn<>'' then
    begin
        cc.Info('getsources: Reading pixel weight image '+wfn);
        SImg_ReadImageFile(cc,wfn,fw);
    end;

    { sources }
    pl.pt:=nil;
    cc.Info('getsources: Finding sources using ' +int2str(cc.threads) + ' threads');
    SImg_GetSources(cc, fi,fw, pl, n,cmp);

    { write output  }
    for n:=0 to pl.ny-1 do
    begin
        pl.ptd[n*pl.sl  ]-=xc;
        pl.ptd[n*pl.sl+1]-=yc;
        end;
    mi.deleteRaw;
    mi.add('IBNDXMIN',real2str(-xc,-1),'Lower image boundary in X direction');
    mi.add('IBNDXMAX',real2str(fi.nx-1-xc,-1),'Upper image boundary in X direction');
    mi.add('IBNDYMIN',real2str(-yc,-1),'Lower image boundary in Y direction');
    mi.add('IBNDYMAX',real2str(fi.ny-1-yc,-1),'Upper image boundary in Y direction');

    cc.Info('getsources: Writing source data array '+ofn);
    SImg_WFITS(cc,CSImg_FileOutput.init(ofn),pl,mi).close(cc);
    if (fw.pt<>nil) then SImg_Free(cc,fw);

    SImg_Free(cc,fi);
    SImg_Free(cc,pl);
    mi.destroy;
    cc.destroy;
end.
