{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name maxset}
{@description Finds the brightest n values of each pixels and return their average.}
{@longdescription
Brightest value can be discarded using option -d (e.g. in order to filter comics).
Then average of next n-1 brightest values is returned.}

{$I directives.pas}
uses simage,getparam;
var par        : tparams;
    cc         : CSImg_Config;
    fi,fo      : ansistring;
    iset       : CSImg_ImageSet;
    img        : TSImg_FPImage;
    mi         : CSImg_MetaInfo;
    n          : longint;
    d          : boolean;

begin
    with par do
    begin
        init;
        paramusage:='<infiles> [<outfile>]';
        fi:='';      defparam('<infiles>','Input images',@fi,tpastr,rapdreq);
        fo:='';      defparam('<outfile>','Output image (default: stdout)',fo);
                     defparam('-h','Help',helpf);
        n:=2;        defparam('-n','Number of brightest values n (default: 2, valid: 1..8)',n);
        d:=false;    defparam('-d','Discard brightest value',d);
        run;
        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('maxset: src='+fi+' out='+fo+' n='+int2str(n)+' d='+boolean2str(d));

    iset:=CSImg_ImageSet_Files.init(cc,fi);
    SImg_MaxSet(cc,iset,img,n,d);

    cc.Info('avgset: Writing output image '+fo);
    mi:=CSImg_MetaInfo.create;
    iset.getMI0(cc,mi);
    SImg_WriteImageFile(cc,fo,img,mi);
    SImg_Free(cc,img);
    iset.done;
    mi.destroy;
    cc.destroy;
end.
