{  simg -- command line applications for deterministic image processing
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name polar}
{@description Polar coordinate transformations}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                : tparams;
    cc                 : CSImg_Config;
    img,ref            : TSImg_FPImage;
    fi,fo,fref         : ansistring;
    inv,fc,fm          : boolean;
    da,dr,xc,yc,a1,a2  : double;
    nx,ny,i            : longint;
    mi                 : CSImg_MetaInfo;

begin
    img.pt:=nil; ref.pt:=nil;
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fi:='';     defparam('<infile>','Source image (default: stdin)',fi);
        fo:='';     defparam('<outfile>','Destination image (default: stdout)',fo);
                    defparam('-h','Help',helpf);
        inv:=false; defparam('-inv','Polar --> Cartesian mode  (default: Cartesian --> polar mode )',inv);
        da:=0.2;    defparam('-da','Differential angle, default: 0.2',da);
        dr:=0.4;    defparam('-dr','Differential radius, default: 0.25',dr);
        xc:=0;      defparam('-xc','X-Center, default: 0',xc);
        yc:=0;      defparam('-yc','Y-Center, default: 0',yc);
        fc:=false;  defparam('-fc','Find center',fc);
        fm:=false;  defparam('-fm','Find maximum',fm);
        a1:=-90;    defparam('-a1','Start angle in degree (default: -90)',a1);
        a2:=450;    defparam('-a2','End angle in degree (default: 450), only for Cartesian --> polar mode',a2);

        fref:='';   defparam('-ref','Reference image, required for examining the center, only for polar --> Cartesian mode',fref);
        nx:=0;      defparam('-nx','X-Size (default: X-size of the reference image), only for Cartesian --> polar mode',nx);
        ny:=0;      defparam('-ny','Y-Size (default: Y-size of the reference image), only for Cartesian --> polar mode',ny);
        run;

        cc:=CSImg_Config.create(SImg_Config_Default);

        cc.debug('polar: values: source='+fi+' dest='+fo+' inv='+boolean2str(inv)+' da='+real2str(da)+' dr='+real2str(dr)+
            ' xc='+real2str(xc)+' yc='+real2str(yc)+' fc='+boolean2str(fc)+' fm='+boolean2str(fm)+' a1='+real2str(a1)+' a2='+real2str(a2)+
            ' fref='+fref+' nx='+int2str(nx)+' ny='+int2str(ny));
        i:=0;
        if vardef(@xc) or vardef(@yc) then i+=1;
        if fc then i+=1;
        if fm then i+=1;
        if i>1 then cc.error('polar: -xc, -yc, -fc, -fm can''t be used together');

        cc.Info('polar: Reading input image '+fi);
        mi:=CSImg_MetaInfo.create;
        SImg_ReadImageFile(cc,fi,img,mi);

        if inv then
        begin
            if vardef(@fref) then
            begin
                cc.Info('polar: Reading reference image '+fref);
                SImg_ReadImageFile(cc,fref,ref)
            end
            else if (not vardef(@nx)) or (not vardef(@ny)) or fc or fm then cc.error('polar: Reference image required');
            if not vardef(@nx) then nx:=ref.nx;
            if not vardef(@ny) then ny:=ref.ny;
            if fc then SImg_FindCenter(cc,ref,xc,yc);
            if fm then SImg_FindMaximum(cc,ref,xc,yc);
            SImg_Free(cc,ref);
            cc.Info('polar: polar -> Cartesian transformation');
            SImg_PolarToCartesian(cc, img,img, nx,ny, xc,yc, dr,da,a1);
        end
        else begin
            if fc then SImg_FindCenter(cc,img,xc,yc);
            if fm then SImg_FindMaximum(cc,img,xc,yc);
            cc.Info('polar: Cartesian ->polar transformation');
            SImg_CartesianToPolar(cc, img,img, xc,yc, dr,da,a1,a2);
        end;

        cc.Info('polar: Writing output image '+fo);
        SImg_WriteImageFile(cc,fo,img,mi);
        SImg_Free(cc,img);

        done;
    end;

    cc.destroy;
end.
