{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*********************************************************************}
{********* SImg_GetColorAdjust ***************************************}
{*********************************************************************}
{ dm    : differential mode
  frad  : smoothing radius
}
function SImg_GetColorAdjust(cc:CSImg_Config; fj:TSImg_FPImage;dm:boolean; frad:double):matrix;
var fi             : TSImg_FPImage;
    a,b0,c         : matrix;
    i,j,k,l,iz,jz  : longint;
    irad,ss        : longint;
    d,e,f          : ^simgfp;
    r              : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetColorAdjust: called with: img='+SImg_InfoStr2(fj)+' dm='+boolean2str(dm)+' frad='+real2str(frad));
    if not SImg_Recover(cc,fj) then cc.error('SImg_GetColorAdjust: Invalid image: fj');

    if fj.nc=1 then
    begin
        SImg_GetColorAdjust:=ident(1);
        SImg_DecDebugLevel;
        exit;
    end;

    fi.pt:=nil;
    irad:=round(frad-0.45);
    if irad<0 then irad:=0;
    ss:=round(sqrt(irad*2+0.3));
    if irad>0 then SImg_Filter(cc,fj,fi,irad+0.5,-1,ss)
    else fi:=fj;

    a.pt:=nil; b0.pt:=nil; c.pt:=nil;
    am(c,zmatrix(fi.nc-1,fi.nc-1));
    am(b0,zmatrix(1,fi.nc-1));
    if dm then
    begin
        iz:=fi.ny-2;
        jz:=fi.nx-2;
    end
    else begin
        iz:=fi.ny-1;
        jz:=fi.nx-1;
    end;
    for i:=0 to iz do
    begin
        d:=@fi.pt[i*fi.sl];
        e:=d+fi.nc;
        f:=d+fi.sl;
        for j:=0 to jz do
        begin
            if dm then for k:=1 to fi.nc-1 do
            begin
                b0.pt[k-1]+=(d[0]-e[0])*(d[k]-e[k]) + (d[0]-f[0])*(d[k]-f[k]);
                for l:=k to fi.nc-1 do
                begin
                    r:=(d[k]-e[k])*(d[l]-e[l]) + (d[k]-f[k])*(d[l]-f[l]);
                    c.pt[(k-1)*c.x+(l-1)]+=r;
                    if l>k then c.pt[(l-1)*c.x+(k-1)]+=r;
                end;
            end
            else
                for k:=1 to fi.nc-1 do
                begin
                    b0.pt[k-1]+=d[0]*d[k];
                    for l:=k to fi.nc-1 do
                    begin
                        r:=d[k]*d[l];
                        c.pt[(k-1)*c.x+(l-1)]+=r;
                        if l>k then c.pt[(l-1)*c.x+(k-1)]+=r;
                    end;
                end;
            d+=fi.nc;
            e+=fi.nc;
            f+=fi.nc;
        end;
    end;
    i:=0;
    for k:=1 to fi.nc-1 do
        for l:=1 to fi.nc-1 do
        begin
            if k=l then c.pt[i]*=(fi.nc-1)
            else c.pt[i]:=-c.pt[i];
            i+=1;
        end;
    am(a,solve(c,b0));
    am2(SImg_GetColorAdjust,zmatrix(1,fi.nc));
    SImg_GetColorAdjust.pt[0]:=1;
    for k:=1 to fi.nc-1 do
        SImg_GetColorAdjust.pt[k]:=a.pt[k-1];
    free(a);
    free(b0);
    free(c);
    SImg_GetColorAdjust.ref:=0;

    if irad>0 then SImg_Free(cc,fi);
    SImg_DecDebugLevel;
end;

{*********************************************************************}
{********* SImg_ColorAdjustInfo **************************************}
{*********************************************************************}
{ dm    : differential mode
  frad  : smoothing radius
  returns an info string
}
function SImg_ColorAdjustInfo(cc:CSImg_Config; fj:TSImg_FPImage; dm:boolean; frad:double):ansistring;
var a     : matrix;
    i     : longint;
begin
    am2(a,SImg_GetColorAdjust(cc,fj,dm,frad));
    result:=real2str(a.pt[0]);
    for i:=1 to fj.nc-1 do
        result+='  '+real2str(a.pt[i]);
    free(a);
    SImg_DecDebugLevel;
end;


{*********************************************************************}
{********* SImg_ColorAdjust ******************************************}
{*********************************************************************}
{ dm    : differential mode
  frad  : smoothing radius
  returns an info string
}
function SImg_ColorAdjust(cc:CSImg_Config; fi:TSImg_FPImage; dm:boolean; frad:double):ansistring;
var a     : matrix;
    i,j   : longint;
    r     : double;
    d     : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetColorAdjust: called with: img='+SImg_InfoStr2(fi) + ' dm='+boolean2str(dm) + ' frad='+real2str(frad) );
    if not SImg_Recover(cc,fi) then cc.error('SImg_ColorAdjust: Invalid image: fi');

    am2(a,SImg_GetColorAdjust(cc,fi,dm,frad));
    result:=real2str(a.pt[0]);
    for i:=1 to fi.nc-1 do
        result+='  '+real2str(a.pt[i]);
    for i:=0 to fi.nc-1 do
    begin
        r:=a.pt[i];
        d:=fi.pt+i;
        for j:=0 to fi.nx*fi.ny-1 do
        begin
            d^*=r;
            d+=fi.nc;
        end;
    end;
    free(a);
    SImg_DecDebugLevel;
end;


{*********************************************************************}
{********* SImg_Saturation01 *****************************************}
{*********************************************************************}
procedure SImg_Saturation01(cc:CSImg_Config; fi:TSImg_FPImage;var fj:TSImg_FPImage; sat,maxsat:double);
var fk          : TSImg_FPImage;
    i,j         : longint;
    r,s,t,u,v,w : double;
    d,e         : ^simgfp;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Saturation01: called with: img='+SImg_InfoStr2(fi)+' img2='+SImg_InfoStr2(fj)+' sat='+real2str(sat)+'  maxsat='+real2str(maxsat));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Saturation01: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    if maxsat<0 then cc.warning('SImg_Saturation01: Invalid maximum saturation: maxsat<0, set it to maxsat=0');
    if maxsat<0.001 then maxsat:=0.001;
    fk.pt:=nil;

    fk.pt:=nil;
    SImg_Duplicate(cc,fi,fk,false);
    s:=1/fi.nc;
    w:=abs(sat);
    d:=fi.pt;
    e:=fk.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        r:=0;
        t:=d^; u:=d^;
        for j:=0 to fi.nc-1 do
        begin
            v:=d[j];
            r+=v;
            if v<t then t:=v;
            if v>u then u:=v;
        end;
        r*=s;
        v:=abs(r)*maxsat;
        if ((u-r)*w>v) or ((r-t)*w>v) then
        begin
            if u-r>r-t then v/=u-r
            else v/=r-t;
        end
        else v:=sat;
        for j:=0 to fi.nc-1 do
        begin
            e^:=r+(d^-r)*v;
            d+=1; e+=1;
        end;
    end;
    SImg_Move(cc, fk,fj);
    SImg_DecDebugLevel;
end;


{*********************************************************************}
{********* SImg_Saturation02 *****************************************}
{*********************************************************************}
procedure SImg_Saturation02(cc:CSImg_Config; fi:TSImg_FPImage;var fj:TSImg_FPImage; sat:double);
var fk      : TSImg_FPImage;
    i,j     : longint;
    r,s     : double;
    t,u     : simgfp;
    d,e     : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    SImg_DebugMessage('SImg_Saturation02: called with: img='+SImg_InfoStr2(fi)+' img2='+SImg_InfoStr2(fj)+' sat='+real2str(sat));
    if not SImg_Recover(cc,fi) then cc.error('SImg_Saturation02: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    fk.pt:=nil;

    sat-=1;
    d:=fi.pt;
    t:=d^; u:=t;
    for i:=0 to fi.size-1 do
    begin
        r:=d^; d+=1;
        if r<t then t:=r;
        if r>u then u:=r;
    end;
    fk.pt:=nil;
    SImg_Duplicate(cc,fi,fk,false);
    d:=fi.pt;
    e:=fk.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        r:=t;
        for j:=0 to fi.nc-1 do
            if d[j]>r then r:=d[j];
        r-=t;
        if r<simg_mina then r:=simg_mina;
        r:=1/r;

        for j:=0 to fi.nc-1 do
        begin
            s:=d^-t; d+=1;
            if s>simg_mina then s*=exp(ln(s*r)*sat)
            else s:=0;
            s+=t;
            if s>u then s:=u;
            e^:=s; e+=1;
        end;
    end;
    SImg_Move(cc, fk,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_SaturationCompensation *************************************}
{*************************************************************************}
procedure SImg_SaturationCompensation(cc:CSImg_Config; fi:TSImg_FPImage);
const max_n = 16;
var i,j,c   : longint;
    r       : simgfp;
    d,e,max : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_IncDebugLevel;
    SImg_DebugMessage('SImg_SaturationCompenstaion: fi='+SImg_InfoStr2(fi));
    if not SImg_Integrity(fi) then cc.error('SImg_SaturationCompenstaion: Invalid image: fi');
    if fi.nc<=1 then exit;

    getmem2(max,max_n*fi.nc*sizeof(simgfp));
    e:=max;
    for c:=0 to fi.nc-1 do
        for i:=0 to max_n-1 do
        begin
            e^:=fi.pt[c];
            e+=1;
        end;

    d:=fi.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        e:=max;
        for c:=0 to fi.nc-1 do
        begin
            if d^>e^ then
            begin
                j:=0;
                e^:=d^;
                while (j<max_n-2) and (e[j]>e[j+1]) do
                begin
                    r:=e[j+1];
                    e[j+1]:=e[j];
                    e[j]:=r;
                    j+=1;
                end;
            end;
            d+=1;
            e+=max_n;
        end;
    end;

{   for i:=0 to fi.nc-1 do
        writeln(i,': ',real2str(max[i*max_n]));}

    r:=max[0];
    for i:=1 to fi.nc-1 do
        if max[i*max_n]<r then r:=max[i*max_n];

    {writeln(real2str(r));}

    d:=fi.pt;
    for i:=0 to fi.size-1 do
        if d[i]>r then d[i]:=r;

    freemem2(max,max_n*fi.nc*sizeof(simgfp));
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Compose ****************************************************}
{*************************************************************************}
{ Color composition using luminance image
  fi        : input colors images
  fo        : output image
}
procedure SImg_Compose(cc:CSImg_Config; fi:array of TSImg_FPImage; var fo:TSImg_FPImage);
var fk          : TSImg_FPImage;
    nc,nc2      : longint;
    i,j,k,c     : longint;
    b           : boolean;
    d,e         : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Compose: called with: fi.size='+int2str(high(fi)+1)+' fo='+SImg_InfoStr2(fo));
    b:=false;
    nc:=0;
    for i:=0 to high(fi) do
    begin
        if not SImg_Recover(cc,fi[i]) then cc.error('SImg_Compose: Invalid image: fi['+int2str(i)+']');
        if (fi[i].nx<>fi[0].nx) or (fi[i].ny<>fi[0].ny) then cc.error('SImg_Compose: Invalid dimensions: fi['+int2str(i)+']');
        b:=b or SImg_OverWriteAlert(fi[i],fo);
        nc+=fi[i].nc;
    end;
    if not b then SImg_Free(cc,fo);

    { adjust fwhm to cfwhm and compose everything }
    fk.pt:=nil;
    SImg_New(cc, fk, fi[0].nx,fi[0].ny,nc);
    j:=0;
    for i:=0 to high(fi) do
    begin
        d:=fk.pt + j;
        e:=fi[i].pt;
        nc2:=fi[i].nc;
        for k:=0 to fk.nx*fk.ny-1 do
        begin
            for c:=0 to nc2-1 do
                d[c]:=e[c];
            d+=nc;
            e+=nc2;
        end;
        j+=nc2;
    end;

    SImg_Move(cc, fk,fo);
    SImg_DecDebugLevel;
end;

{*************************************************************************}
{******* SImg_LCompose ***************************************************}
{*************************************************************************}
{ Color composition using luminance image
  fi        : input colors images
  fj        : input luminance image
  fo        : output image
  fwhm      : FWHM's of input images
  cfwhm     : minimum color FWHM
  result    : color FWHM
}
function SImg_LCompose(cc:CSImg_Config; fi:array of TSImg_FPImage; fj:TSImg_FPImage; var fo:TSImg_FPImage; fwhm:array of double; cfwhm:double):double;
const reg  = 1e-2;      { regularization }
var fk,fl       : TSImg_FPImage;
    r,s,t,u     : double;
    size,nc     : longint;
    i,j,k,c     : longint;
    b           : boolean;
    d,e         : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_LCompose: called with: fi.size='+int2str(high(fi)+1)+' fj='+SImg_InfoStr2(fj)+' fo='+SImg_InfoStr2(fo)+' cfwhm='+real2str(cfwhm));
    if not SImg_Recover(cc,fj) then cc.error('SImg_LCompose: Invalid luminance image');
    if fj.nc<>1 then cc.error('SImg_LCompose: Luminance image must only have one color');

    if high(fi)<>high(fwhm) then cc.error('SImg_LCompose: Number of images must be equal to number of FWHM');
    size:=high(fi)+1;
    if cfwhm<1 then cfwhm:=1;
    b:=false;
    nc:=0;
    for i:=0 to size-1 do
    begin
        if not SImg_Recover(cc,fi[i]) then cc.error('SImg_LCompose: Invalid image: fi['+int2str(i)+']');
        if (fi[i].nx<>fj.nx) or (fi[i].ny<>fj.ny) then cc.error('SImg_LCompose: Invalid dimensions: fi['+int2str(i)+']');
        fwhm[i]:=abs(fwhm[i]);
        if fwhm[i]<1 then fwhm[i]:=1;
        if fwhm[i]>cfwhm then cfwhm:=fwhm[i];
        b:=b or SImg_OverWriteAlert(fi[i],fo);
        nc+=fi[i].nc;
    end;
    if (not b) and (not SImg_OverWriteAlert(fj,fo)) then SImg_Free(cc,fo);
    result:=cfwhm;

    { adjust fwhm to cfwhm and compose everything }
    fk.pt:=nil;
    SImg_New(cc, fk, fj.nx,fj.ny,nc);
    j:=0; s:=0;
    for i:=0 to size-1 do
    begin
        fl.pt:=nil;
        r:=sqr(cfwhm)-sqr(fwhm[i]);
        if r<0.25 then fl:=fi[i]
        else SImg_Filter(cc, fi[i],fl, sqrt(r)*FWHM_TO_RAD);
        d:=fk.pt + j;
        e:=fl.pt;
        for k:=0 to fj.size-1 do
        begin
            for c:=0 to fl.nc-1 do
            begin
                r:=e[c];
                d[c]:=r;
                if r>0 then s+=r;
            end;
            d+=nc;
            e+=fl.nc;
        end;
        j+=fl.nc;
        if fl.pt<>fi[i].pt then SImg_Free(cc,fl);
    end;

    { luminance adjustment }
    if s<1e-30 then cc.warning('SImg_LCompose: non-positive colors')
    else begin
        s:=reg*s/fk.size;
        d:=fk.pt;
        for k:=0 to fj.size-1 do
        begin
            t:=0;
            u:=fj.pt[k];
            if u>0 then
            begin
                for c:=0 to nc-1 do
                begin
                    r:=d[c];
                    if r<0 then r:=0;
                    r+=s;
                    t+=r;
                    d[c]:=r;
                end;
                t:=u/t;
            end;
            for c:=0 to nc-1 do
                d[c]*=t;
            d+=nc;
        end;
    end;

    if b and SImg_OverWriteAlert(fj,fo) then fo:=fk
    else SImg_Move(cc, fk,fo);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_RegBounds **************************************************}
{*************************************************************************}
{ Interpolate undefined regions at boundary by regularization using other colors.
  fi    : input and output images
  wi    : weight images
  fo    : output images. Must be Allocated and can be equal to fi or wi.
  b     : boundary width in pixel
  reg   : regularization factor
}
procedure SImg_RegBounds(cc:CSImg_Config; fi,wi:array of TSImg_FPImage; b:longint; reg:double);
var nx,ny,ni,mx,my      : longint;
    i,x,y,c             : longint;
    r,s,sfw,sw,yr,br    : double;
    d,e                 : ^simgfp;
    yb                  : boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_RegBounds called with: fi.size='+int2str(high(fi)+1)+' wi.size='+int2str(high(fi)+1)+' b='+int2str(b)+' reg='+real2str(reg));
    if high(fi)<>high(wi) then cc.error(cc.name+': Number of given input images, weights and output images must be equal');
    ni:=high(fi);
    nx:=fi[0].nx;  ny:=fi[0].ny;
    for i:=0 to ni do
    begin
        if not SImg_Recover(cc,fi[i]) then cc.error('SImg_RegBounds: Invalid image: fi['+int2str(i)+']');
        if not SImg_Recover(cc,wi[i]) then cc.error('SImg_RegBounds: Invalid image: wi['+int2str(i)+']');
        if (fi[i].nx<>nx) or (fi[i].ny<>ny) then cc.error(cc.name+': Invalid dimension for input image '+int2str(i));
        if (wi[i].nc<>fi[i].nc) or (wi[i].nx<>nx) or (wi[i].ny<>ny) then cc.error(cc.name+': Invalid dimension for weight '+int2str(i));
    end;
    if b<1 then b:=1;
    if b>nx div 2 then b:=nx div 2;
    if b>ny div 2 then b:=ny div 2;
    if reg<simg_mina then reg:=simg_mina;

    br:=reg/(2*b);
    ny-=1;  my:=2*b-ny;
    nx-=1;  mx:=2*b-nx;
    for y:=0 to ny-1 do
    begin
        yr:=(my+abs(2*y-ny))*br;
        yb:=yr>0;
        x:=0;
        repeat
            { sums }
            sfw:=0; sw:=0;
            for i:=0 to ni do
            begin
                c:=y*fi[i].sl + x*fi[i].nc;
                d:=fi[i].pt + c;
                e:=wi[i].pt + c;
                for c:=0 to fi[i].nc-1 do
                begin
                    r:=e[c];
                    sfw+=d[c]*r;
                    sw+=r;
                end;
            end;
            { regularization factor }
            r:=(mx+abs(2*x-nx))*br;
            if yr>r then r:=yr;
            sfw*=r;
            sw*=r;
            { regularized result }
            for i:=0 to ni do
            begin
                c:=y*fi[i].sl + x*fi[i].nc;
                d:=fi[i].pt + c;
                e:=wi[i].pt + c;
                for c:=0 to fi[i].nc-1 do
                begin
                    r:=e[c];
                    s:=r+sw;
                    if s>simg_mina then d[c]:=(d[c]*r+sfw)/s;
                end;
            end;
            { next pixel }
            x+=1;
            if (not yb) and (x=b) then x:=nx-b+1;
        until x>nx;
    end;

    SImg_DecDebugLevel;
end;
