{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* FITS (image format) routines ************************************}
{*************************************************************************}
{*************************************************************************}
type TSImg_FITSCard=packed record
       key  : array[0..8]  of char;
       val  : array[0..70] of char;
       end;

procedure SImg_FITS_getval(const c:TSImg_FITSCard; var value,comment:string71);
var i,j : longint;
    q   : boolean;
begin
    value:='';
    i:=0;
    while (i<=70) and (c.val[i]<=#32)  do
        inc(i);
    q:=false;
    while (i<=70) and (q or ((c.val[i]<>'/') and (c.val[i]>#32) and (c.val[i]<#128))) do
    begin
        value+=c.val[i];
        if c.val[i]='''' then q:=not q;
        inc(i);
    end;
    j:=length(value);
    while (j>=0) and (value[j]=#32) do
        dec(j);
    value[0]:=char(j);

    { comment }
    comment:='';
    while (i<70) and (c.val[i]<>'/') do
        inc(i);
    inc(i);
    while (i<70) and (c.val[i]<=#32)  do
        inc(i);
    j:=70;
    while (j>=i) and (c.val[j]<=#32) do
        dec(j);
    if j>=i then
    begin
        move(c.val[i],comment[1],j-i+1);
        comment[0]:=char(j-i+1);
    end;
end;

function SImg_FITS_getkey(const c:TSImg_FITSCard):string8;
var i : longint;
begin
    i:=0;
    while (i<=7) and (c.key[i]<=#32) do
        inc(i);
    result:='';
    while (i<=7) and (c.key[i]>#32) and (c.key[i]<#128) do
    begin
        result+=c.key[i];
        inc(i);
    end;
end;

function SImg_RFITS(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo; headerOnly:boolean):boolean;
var he               : ^TSimg_FITSCARD;
    key              : string8;
    value,comment    : string71;
    bole             : boolean;
    fmt              : TSImg_RawFormat;
    x,y,c,bpp        : longint;
    cf,co,r          : double;
    i                : longint;
    b                : array[0..3] of byte absolute i;

@rawptr d;
@rawptr e;

begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_RFITS: called with: ri='+SImg_InfoStr2(ri)+'   '+int2str(fb.size)+' Bytes input buffer');
    SImg_Free(cc,ri);
    SImg_RFITS:=false;
    he:=@fb.pt[0];
    SImg_FITS_getval(he[0],value,comment);
    if (SImg_FITS_getkey(he[0])<>'SIMPLE') or (((fb.size<=2880) or (value<>'T')) and (not headerOnly)) then
        exit;

    bole:=false;
    ri.nx:=-1;  ri.ny:=-1;  ri.nc:=-1;
    co:=0;  cf:=1;
    bpp:=0;
    i:=1;
    repeat
        key:=SImg_FITS_getkey(he[i]);
        SImg_FITS_getval(he[i],value,comment);
        if he[i].key[8]='=' then
        begin
            if key='BITPIX' then
            begin
                val(value,x,y);
                if (y=0) and ((x=-64) or (x=-32) or (x=8) or (x=16) or (x=32)) then bpp:=x;
            end
            else if key='NAXIS' then
            begin
                val(value,x,y);
                if (y=0) and (x=2) then ri.nc:=1;
            end
            else if key='NAXIS1' then
            begin
                val(value,x,y);
                if (y=0) and (x>1) then ri.nx:=x;
            end
            else if key='NAXIS2' then
            begin
                val(value,x,y);
                if (y=0) and (x>1) then ri.ny:=x;
            end
            else if key='NAXIS3' then
                begin
                val(value,x,y);
                if (y=0) and (x>0) then ri.nc:=x;
            end
            else if key='BZERO' then
            begin
                val(value,r,y);
                if y=0 then co:=r;
            end
            else if key='BSCALE' then
            begin
                val(value,r,y);
                if (y=0) and (abs(r)>1e-20) then cf:=r;
            end
            else if key='BYTEORDR' then
            begin
                if value='''LITTLE_ENDIAN''' then bole:=true
                else if value='''BIG_ENDIAN''' then bole:=false;
            end
            else if mi<>nil then mi.add(key,value,comment);
        end;
        inc(i);
    until (key='END') or (i*80>fb.size);

    if headerOnly then
    begin
        SImg_RFITS:=true;
        exit;
    end;

    if (ri.nx<2) or (ri.ny<2) or (ri.nc<1) or (bpp=0) then
    begin
        SImg_Warning('SImg_RFITS: Incomplete primary header information:');
        if ri.nx<2 then SImg_Warning('SImg_RFITS: key NAXIS1 is missed or invalid');
        if ri.ny<2 then SImg_Warning('SImg_RFITS: key NAXIS2 is missed or invalid');
        if ri.nc<1 then SImg_Warning('SImg_RFITS: key NAXIS3 is missed or invalid or `NAXIS   = 2''');
        if bpp=0 then cc.Warning('SImg_RFITS: keyword BITPIX is missed or invalid');
        exit;
    end;

    i:=(((i-1) div 36)+1)*36;
    if fb.size<ri.nx*ri.ny*ri.nc*(abs(bpp) div 8) + i*80 then
    begin
        cc.Warning('SImg_RFITS: Unexpected end of file');
        exit;
    end;

    SImg_RFITS:=true;
    case bpp of
        -64    : fmt:=SImg_RawDouble;
        -32    : fmt:=SImg_RawSingle;
        16     : fmt:=SImg_RawWord;
        8      : fmt:=SImg_RawByte;
        else fmt:=SImg_RawFP;
    end;
    SImg_New(cc, ri, fmt, ri.nx, ri.ny, ri.nc, false);
    ri.cf:=cf; ri.co:=co;
    if fmt=SImg_RawWord then ri.co-=32768*cf;
    if bpp=32 then
    begin
        ri.co:=0;
        ri.cf:=1;
    end;

    d:=@fb.pt[i*80];
    if bole then bpp+=1;
    case bpp of
        -64 :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        ed:=@ri.ptd[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            e[0]:=d[7];
                            e[1]:=d[6];
                            e[2]:=d[5];
                            e[3]:=d[4];
                            e[4]:=d[3];
                            e[5]:=d[2];
                            e[6]:=d[1];
                            e[7]:=d[0];
                            ed+=ri.nc; dd+=1;
                            end;
                        end;
        -63 :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        ed:=@ri.ptd[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            ed^:=dd^;
                            ed+=ri.nc;  dd+=1;
                        end;
                    end;
        -32 :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        es:=@ri.pts[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            e[0]:=d[3];
                            e[1]:=d[2];
                            e[2]:=d[1];
                            e[3]:=d[0];
                            es+=ri.nc; ds+=1;
                        end;
                    end;
        -31 :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        es:=@ri.pts[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            es^:=ds^;
                            es+=ri.nc;  ds+=1;
                        end;
                    end;
        32  :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        ep:=@ri.ptp[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            b[0]:=d[3];
                            b[1]:=d[2];
                            b[2]:=d[1];
                            b[3]:=d[0];
                            ep^:=i*cf+co;
                            ep+=ri.nc; ds+=1;
                        end;
                    end;
        33  :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        ep:=@ri.ptp[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            ep^:=(plongint(ds))^*cf+co;
                            ep+=ri.nc; ds+=1;
                        end;
                    end;
        16  :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        ew:=@ri.ptw[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            e[0]:=d[1];
                            e[1]:=d[0] xor 128;
                            ew+=ri.nc; dw+=1;
                        end;
                    end;
        17  :   for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        ew:=@ri.ptw[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            ew^:=dw^ xor 32768;
                            ew+=ri.nc;  dw+=1;
                        end;
                    end;
        else    for c:=0 to ri.nc-1 do
                    for y:=0 to ri.ny-1 do
                    begin
                        eb:=@ri.ptb[(ri.ny-1-y)*ri.sl+c];
                        for x:=0 to ri.nx-1 do
                        begin
                            eb^:=db^;
                            eb+=ri.nc; db+=1;
                        end;
                    end;
    end;
end;

function SImg_RFITS(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
begin
    result:=SImg_RFITS(cc,fb,ri,mi,false);
end;

{*************************************************************************}
{******* SImg_WFITS ******************************************************}
{*************************************************************************}
procedure SImg_FITS_cardset(var card:TSImg_FITSCard; const key:string8; const val,comment:ansistring);
var i,j  : longint;
begin
{ key }
    for i:=1 to length(key) do
        card.key[i-1]:=key[i];
{ value }
    j:=0;
    if val<>'' then
    begin
        card.key[8]:='=';
        for i:=1 to length(val) do
            card.val[i]:=val[i];
        j:=length(val)+1;
    end;
{ comment }
    if (j<=67) and (comment<>'') then
    begin
        card.val[j+1]:='/';
        for i:=1 to length(comment) do
            if j+2+i<=70 then card.val[j+2+i]:=comment[i];
    end;
end;

{ returns f, does not close it }
function SImg_WFITS(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo):CSImg_Output;
var ch        : ^TSImg_FITSCard;
    x,y,c,bpp : longint;
    bs        : int64;
    bole      : boolean;
    co        : double;
@rawptr d;
@rawptr e;
@rawptr buf;
begin
    result:=f;
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_WFITS: called with: output='+f.info+' ri='+SImg_InfoStr2(ri));
    if (ri.pt<>nil) and (not SImg_Recover(cc,ri)) then
    begin
        cc.warning('SImg_WFITS: Invalid image: only writing header');
        ri.pt:=nil;
    end;

    bole:=(cc.forceLE) and (not cc.forceBE);
    if ri.pt<>nil then
    begin
        if not ri.bole then SImg_SwapBytes(ri);
        if ri.nc>1 then y:=10
        else y:=9;
    end
    else y:=4;

    if mi<>nil then
        for x:=0 to mi.size-1 do
            if mi.key[x]<>'' then y+=1;
    c:=((y+35) div 36)*2880;
    getmem2(ch,c);
    fillchar(ch^,c,#32);
    SImg_FITS_cardset(ch[0],'SIMPLE','T','');

    if ri.pt=nil then bpp:=8
    else if ri.fmt=SImg_RawWord then bpp:=16
    else if ri.fmt=SImg_RawSingle then bpp:=-32
    else if ri.fmt=SImg_RawDouble then bpp:=-64
    else bpp:=8;
    SImg_FITS_cardset(ch[1],'BITPIX',int2str(bpp),'');

    if ri.pt=nil then x:=0
    else if ri.nc=1 then x:=2
    else x:=3;
    SImg_FITS_cardset(ch[2],'NAXIS',int2str(x),'');
    x+=3;

    if ri.pt<>nil then
    begin
        SImg_FITS_cardset(ch[3],'NAXIS1',int2str(ri.nx),'');
        SImg_FITS_cardset(ch[4],'NAXIS2',int2str(ri.ny),'');
        if x=6 then SImg_FITS_cardset(ch[5],'NAXIS3',int2str(ri.nc),'');
        co:=ri.co;
        if ri.fmt=SImg_RawWord then co+=32768*ri.cf;
        SImg_FITS_cardset(ch[x+0],'BZERO',real2str(co,15),'');
        SImg_FITS_cardset(ch[x+1],'BSCALE',real2str(ri.cf,15),'');
        SImg_FITS_cardset(ch[x+2],'BYTEORDR',condstring(bole,'''LITTLE_ENDIAN''','''BIG_ENDIAN'''),'');
        x+=3;
    end;

    if mi<>nil then
        for y:=0 to mi.size-1 do
            if mi.key[y]<>'' then
            begin
                SImg_FITS_cardset(ch[x], mi.key[y], mi.val[y], mi.comment[y]);
                inc(x);
            end;

    SImg_FITS_cardset(ch[x],'END','','');

    f.open(cc);
    f.write(cc,ch,c);
    freemem2(ch);

    if ri.pt=nil then exit;

    bs:=ri.nx*ri.ny*(abs(bpp)>>3);
    getmem2(buf,bs);
    if bole then bpp+=1;

    case bpp of
        -64 :   for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        dd:=ri.ptd + (y*ri.sl+c);
                        ed:=bufd + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            e[0]:=d[7];
                            e[1]:=d[6];
                            e[2]:=d[5];
                            e[3]:=d[4];
                            e[4]:=d[3];
                            e[5]:=d[2];
                            e[6]:=d[1];
                            e[7]:=d[0];
                            dd+=ri.nc; ed+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                end;
        -63 :   for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        dd:=ri.ptd + (y*ri.sl+c);
                        ed:=bufd + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            ed^:=dd^;
                            dd+=ri.nc; ed+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                end;
        -32 :   for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        ds:=ri.pts + (y*ri.sl+c);
                        es:=bufs + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            e[0]:=d[3];
                            e[1]:=d[2];
                            e[2]:=d[1];
                            e[3]:=d[0];
                            ds+=ri.nc; es+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                 end;
        -31 :   for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        ds:=ri.pts + (y*ri.sl+c);
                        es:=bufs + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            es^:=ds^;
                            ds+=ri.nc; es+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                end;
        16  :   for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        dw:=ri.ptw + (y*ri.sl+c);
                        ew:=bufw + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            e[0]:=d[1] xor 128;
                            e[1]:=d[0];
                            dw+=ri.nc; ew+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                end;
        17  :   for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        dw:=ri.ptw + (y*ri.sl+c);
                        ew:=bufw + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            ew^:=dw^ xor 32786;
                            dw+=ri.nc; ew+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                end;
        else    for c:=0 to ri.nc-1 do
                begin
                    for y:=0 to ri.ny-1 do
                    begin
                        db:=ri.ptb + (y*ri.sl+c);
                        eb:=buf + (ri.ny-1-y)*ri.nx;
                        for x:=0 to ri.nx-1 do
                        begin
                            eb^:=db^;
                            db+=ri.nc; eb+=1;
                        end;
                    end;
                    f.write(cc,buf,bs);
                end;
    end;

    x:=(bs*ri.nc) mod 2880;
    if x>0 then
    begin
        if bs<2880-x then
        begin
            freemem2(buf,bs);
            bs:=2880-x;
            getmem2(buf,bs);
        end;
        fillchar(buf^,2880-x,#0);
        f.write(cc,buf,2880-x);
    end;

    freemem2(buf,bs);
end;

{*************************************************************************}
{******* SImg_WFITSHeader ************************************************}
{*************************************************************************}
{ returns f, does not close it }
function SImg_WFITSHeader(cc:CSImg_Config; f:CSImg_Output; mi:CSImg_MetaInfo):CSImg_Output;
var ri  : TSImg_RDImage;
begin
    SImg_IncDebugLevel;
    result:=f;
    ri.pt:=nil;
    SImg_WFITS(cc,f,ri,mi);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_AFITSTable *************************************************}
{*************************************************************************}
{ returns f, does not close it }
function SImg_AFITSTable(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; const header:array of string8; mi:CSImg_MetaInfo):CSImg_Output;
var ch          : ^TSImg_FITSCard;
    x,y,c,b,nx  : longint;
    bs          : int64;
    bole        : boolean;
    fmt         : string8;
@rawptr d;
@rawptr e;
@rawptr buf;
begin
    result:=f;
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_WFITSTable: called with: output='+f.info+' ri='+SImg_InfoStr2(ri));

    if (ri.pt<>nil) and (not SImg_Recover(cc,ri)) then cc.error('No data array given');
    if ri.nc<>1 then cc.error('SImg_WFITS: Data array must be 2 dimensional (nc must be 1)');

    nx:=high(header)+1;
    if ri.nx<nx then nx:=ri.nx;
    if nx<1 then cc.warning('SImg_Error: No columns given');

    if ri.fmt=SImg_RawWord then b:=2
    else if ri.fmt=SImg_RawSingle then b:=4
    else if ri.fmt=SImg_RawDouble then b:=8
    else b:=1;
    if ri.fmt=SImg_RawWord then fmt:='''1I'''
    else if ri.fmt=SImg_RawSingle then fmt:='''1E'''
    else if ri.fmt=SImg_RawDouble then fmt:='''1D'''
    else fmt:='''1B''';

    y:=10;
    if mi<>nil then
        for x:=0 to mi.size-1 do
            if mi.key[x]<>'' then y+=1;

    bole:=(cc.forceLE) and (not cc.forceBE);
    c:=((y+nx*2+35) div 36)*2880;
    getmem2(ch,c);
    fillchar(ch^,c,#32);
    SImg_FITS_cardset(ch[0],'XTENSION','''BINTABLE''','');
    SImg_FITS_cardset(ch[1],'BITPIX','8','');
    SImg_FITS_cardset(ch[2],'NAXIS','2','');
    SImg_FITS_cardset(ch[3],'NAXIS1',int2str(b*nx),'Bytes per row');
    SImg_FITS_cardset(ch[4],'NAXIS2',int2str(ri.ny),'Number of rows');
    SImg_FITS_cardset(ch[5],'TFIELDS',int2str(nx),'Number of columns');
    SImg_FITS_cardset(ch[6],'PCOUNT','0','No special data');
    SImg_FITS_cardset(ch[7],'GCOUNT','1','One single table');
    SImg_FITS_cardset(ch[8],'BYTEORDR',condstring(bole,'''LITTLE_ENDIAN''','''BIG_ENDIAN'''),'');

    for x:=1 to nx do
    begin
        SImg_FITS_cardset(ch[7+x*2  ],'TTYPE'+int2str(x),''''+header[x-1]+'''','Label for column '+int2str(x));
        SImg_FITS_cardset(ch[7+x*2+1],'TFORM'+int2str(x),fmt,'Format for column '+int2str(x));
    end;

    x:=9+nx*2;
    if mi<>nil then
        for y:=0 to mi.size-1 do
            if mi.key[y]<>'' then
            begin
                SImg_FITS_cardset(ch[x], mi.key[y], mi.val[y], mi.comment[y]);
                x+=1;
            end;

    SImg_FITS_cardset(ch[x],'END','','');

    f.write(cc,ch,c);
    freemem2(ch);

    bs:=nx*ri.ny*b;
    getmem2(buf,bs);
    if bole then b+=10;
    case b of
        8 :     for y:=0 to ri.ny-1 do
                begin
                    dd:=ri.ptd + y*ri.sl;
                    ed:=bufd + y*nx;
                    for x:=1 to nx do
                    begin
                        e[0]:=d[7];
                        e[1]:=d[6];
                        e[2]:=d[5];
                        e[3]:=d[4];
                        e[4]:=d[3];
                        e[5]:=d[2];
                        e[6]:=d[1];
                        e[7]:=d[0];
                        dd+=1; ed+=1;
                    end;
                end;
        18 :    for y:=0 to ri.ny-1 do
                    begin
                    dd:=ri.ptd + y*ri.sl;
                    ed:=bufd + y*nx;
                    for x:=1 to nx-1 do
                    begin
                        ed^:=dd^;
                        dd+=1; ed+=1;
                    end;
                end;
        4 :     for y:=0 to ri.ny-1 do
                begin
                    ds:=ri.pts + y*ri.sl;
                    es:=bufs + y*nx;
                    for x:=1 to nx do
                    begin
                        e[0]:=d[3];
                        e[1]:=d[2];
                        e[2]:=d[1];
                        e[3]:=d[0];
                        ds+=1; es+=1;
                    end;
                 end;
        14 :    for y:=0 to ri.ny-1 do
                begin
                    ds:=ri.pts + y*ri.sl;
                    es:=bufs + y*nx;
                    for x:=0 to nx-1 do
                    begin
                        es^:=ds^;
                        ds+=1; es+=1;
                    end;
                end;
        2  :    for y:=0 to ri.ny-1 do
                begin
                    dw:=ri.ptw + y*ri.sl;
                    ew:=bufw + y*nx;
                    for x:=0 to nx-1 do
                    begin
                        e[0]:=d[1] xor 128;
                        e[1]:=d[0];
                        dw+=1; ew+=1;
                    end;
                end;
        12  :   for y:=0 to ri.ny-1 do
                begin
                    dw:=ri.ptw + y*ri.sl;
                    ew:=bufw + y*nx;
                    for x:=0 to nx-1 do
                    begin
                        ew^:=dw^ xor 32786;
                        dw+=1; ew+=1;
                    end;
                end;
        else    for y:=0 to ri.ny-1 do
                begin
                    db:=ri.ptb + y*ri.sl;
                    eb:=buf + y*nx;
                    for x:=0 to nx-1 do
                    begin
                        eb^:=db^;
                        db+=1; eb+=1;
                    end;
                end;
    end;
    f.write(cc,buf,bs);

    x:=bs mod 2880;
    if x>0 then
    begin
        if bs<2880-x then
        begin
            freemem2(buf,bs);
            bs:=2880-x;
            getmem2(buf,bs);
        end;
        fillchar(buf^,2880-x,#0);
        f.write(cc,buf,2880-x);
    end;

    freemem2(buf,bs);
    SImg_DecDebugLevel;
end;
