{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SIMG_FFTTest ****************************************************}
{*************************************************************************}
{ Benchmark:
   img   : test image
   pad   : allow padding
   it    : longint
   pr    : kpixel/s
   flops : KFLOP/s
   t     : time spent }
procedure SImg_FFTTest(cc:CSImg_Config; img:TSImg_FPImage; pad:boolean; it:longint; var pr,flops,t:double);
var i,k    : longint;
    fi     : TSImg_FFTImage;
    ti     : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FFTTest: called with: img='+SImg_InfoStr2(img)+' pad='+boolean2str(pad)+' it='+int2str(it));
    if not SImg_Recover(cc,img) then cc.Error('SImg_FFTTest: Invalid image');
    fi.opt:=nil;

    if pad then SImg_New_findbase(cc,fi,img.nx,img.ny)
    else SImg_New(cc,fi,img.nx,img.ny);

    t:=0;
    for k:=0 to img.nc-1 do
    begin
        if k>0 then SImg_Clear(cc,fi);
        SImg_Copy(cc,img,fi,k);
        ti:=stimer;
        for i:=0 to it-1 do
        begin
            SImg_FFT_r2c(cc,fi);
            SImg_FFT_c2r(cc,fi);
        end;
        t+=stimer-ti;
        SImg_Copy(cc,fi,img,k);
    end;
    pr:=fi.nxr*fi.ny*img.nc*2*it/(t*1000);    { kpixel/s }
    flops:=img.nc*it/(t*1000)*(SImg_FFT_r2c_flops(cc,fi)+SImg_FFT_c2r_flops(cc,fi));    { KFLOPS/s }

    SImg_Free(cc,fi);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_Spectrum ***************************************************}
{*************************************************************************}
{ compute Fourier transform of an image and outputs the spectrum.
    fi   : Source
    fj   : Destination
    cent : True: (0,0) in center
           False: (0,0) top left
   pad   : allow padding}
procedure SImg_Spectrum(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; cent:boolean; pad:boolean);
var ffi         : TSImg_FFTImage;
    fk          : TSImg_FPImage;
    i,j,k,y     : longint;
    x,x0        : longint;
    d           : ^simgfp;
    e           : ^fftfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FFT: called with: source='+SImg_InfoStr2(fi)+' dest='+SImg_InfoStr2(fj)+' pad='+boolean2str(pad)+' center='+boolean2str(cent));
    if not SImg_Integrity(fi) then cc.error('SImg_FFT: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);

    ffi.opt:=nil; fk.pt:=nil;
    if pad then SImg_New_findbase(cc,ffi,fi.nx,fi.ny)
    else SImg_New(cc,ffi,fi.nx,fi.ny);

    SImg_New(cc,fk, ffi.nxr, ffi.ny, fi.nc);
    if cent then
    begin
        x0:=(fk.nx-1) shr 1;
        y:= (fk.ny-1) shr 1;
    end
    else begin
        x0:=0;
        y:=0;
    end;
    for k:=0 to fi.nc-1 do
    begin
        if k>0 then SImg_Clear(cc,ffi);
        SImg_Copy(cc,fi,ffi,k);
        SImg_FFT_r2c(cc,ffi);
        d:=fk.pt + k;
        for i:=0 to fk.ny-1 do
        begin
            x:=x0;
            for j:=0 to fk.nx-1 do
            begin
                if x<ffi.nxc then e:=ffi.pt + y*ffi.nxc*2 + x*2
                else e:=ffi.pt + y*ffi.nxc*2 + (ffi.nxr-x)*2;
                d^:=sqrt(sqr(e[0])+sqr(e[1]));
                d+=fk.nc;
                x+=1;
                if x>=fk.nx then x:=0;
            end;
            y+=1;
            if y>=fk.ny then y:=0;
        end;
    end;
    SImg_Free(cc,ffi);
    SImg_Move(cc,fk,fj);
    SImg_DecDebugLevel;
end;

