{  simg -- image processing library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_Clear ******************************************************}
{*************************************************************************}
procedure SImg_Clear(cc:CSImg_Config; fi:TSImg_FPImage; k:longint);                         { fi(*,*,l):=0; }
var i  : longint;
    d  : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Clear: illegal image');
    if (k<0) or (k>=fi.nc) then cc.error('SImg_Clear: illegal color of fh');
    d:=@fi.pt[k];
    for i:=0 to fi.nx*fi.ny-1 do
        d[i*fi.nc]:=0;
end;

procedure SImg_Clear(cc:CSImg_Config; fi:TSImg_FPImage);                                    { fi(*,*,*):=0; }
var i  : longint;
    d  : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Clear: illegal image');
    d:=fi.pt;
    for i:=0 to fi.size-1 do
        d[i]:=0;
end;


{*************************************************************************}
{******* SImg_Copy *******************************************************}
{*************************************************************************}
procedure SImg_Copy(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint;xt,yt:longint);   { fi(*,*,l):=0; copy fh(y-yt,x-xt,k) --> fi(y,x,l), shift & cut }
var d,e             : ^simgfp;
    i,i1,i2,j,j1,j2 : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fh) then cc.error('SImg_Copy: invalid image: fh');
    if (k<0) or (k>=fh.nc) then cc.error('SImg_Copy: illegal color of fh');
    if not SImg_Recover(cc,fi) then cc.error('SImg_Copy: illegal image: fi');
    if (l<0) or (l>=fi.nc) then cc.error('SImg_Copy: illegal color of fi');
    if (fh.nx<fi.nx) or (fh.ny<fi.ny) then
    begin
        d:=@fi.pt[l];
        for i:=0 to fi.nx*fi.ny-1 do
            d[i*fi.nc]:=0;
    end;
    i1:=max(0,yt); i2:=min(fi.ny,fh.ny+yt)-1;
    j1:=max(0,xt); j2:=min(fi.nx,fh.nx+xt)-1;
    for i:=i1 to i2 do
    begin
        d:=@fh.pt[(i-yt)*fh.sl+ (j1-xt)*fh.nc + k];
        e:=@fi.pt[i*fi.sl + j1*fi.nc + l];
        for j:=j1 to j2 do
        begin
            e^:=d^;
            d+=fh.nc; e+=fi.nc;
        end;
    end;
end;

procedure SImg_Copy(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint);                 { fi(*,*,l):=0; copy fh(y   ,x   ,k) --> fi(y,x,l), cut }
begin
    SImg_Copy(cc,fh,fi,k,l,0,0);
end;


{*************************************************************************}
{******* SImg_Add ********************************************************}
{*************************************************************************}
procedure SImg_Add(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint;xt,yt:longint);   { fi(y,x,l)+=fh(y-yt,x-xt,k), shift & cut }
var d,e             : ^simgfp;
    i,i1,i2,j,j1,j2 : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fh) then cc.error('SImg_Add: invalid image: fh');
    if (k<0) or (k>=fh.nc) then cc.error('SImg_Add: illegal color of fh');
    if not SImg_Recover(cc,fi) then cc.error('SImg_Add: illegal image: fi');
    if (l<0) or (l>=fi.nc) then cc.error('SImg_Add: illegal color of fi');
    i1:=max(0,yt); i2:=min(fi.ny,fh.ny+yt)-1;
    j1:=max(0,xt); j2:=min(fi.nx,fh.nx+xt)-1;
    for i:=i1 to i2 do
    begin
        d:=@fh.pt[(i-yt)*fh.sl+ (j1-xt)*fh.nc + k];
        e:=@fi.pt[i*fi.sl + j1*fi.nc + l];
        for j:=j1 to j2 do
        begin
            e^+=d^;
            d+=fh.nc; e+=fi.nc;
        end;
    end;
end;

procedure SImg_Add(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint);                 {  fi(y,x,l)+=fh(y   ,x   ,k), cut }
begin
    SImg_Add(cc,fh,fi,k,l,0,0);
end;


{*************************************************************************}
{******* SImg_Mul ********************************************************}
{*************************************************************************}
procedure SImg_Mul(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint;xt,yt:longint);   { fi(y,x,l)*=fh(y-yt,x-xt,k), shift & cut, undefined pixels assumed 1 }
var d,e             : ^simgfp;
    i,i1,i2,j,j1,j2 : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fh) then cc.error('SImg_Mul: invalid image: fh');
    if (k<0) or (k>=fh.nc) then cc.error('SImg_Mul: illegal color of fh');
    if not SImg_Recover(cc,fi) then cc.error('SImg_Mul: illegal image: fi');
    if (l<0) or (l>=fi.nc) then cc.error('SImg_Mul: illegal color of fi');
    i1:=max(0,yt); i2:=min(fi.ny,fh.ny+yt)-1;
    j1:=max(0,xt); j2:=min(fi.nx,fh.nx+xt)-1;
    for i:=i1 to i2 do
    begin
        d:=@fh.pt[(i-yt)*fh.sl+ (j1-xt)*fh.nc + k];
        e:=@fi.pt[i*fi.sl + j1*fi.nc + l];
        for j:=j1 to j2 do
        begin
            e^*=d^;
            d+=fh.nc; e+=fi.nc;
        end;
    end;
end;

procedure SImg_Mul(cc:CSImg_Config; fh,fi:TSImg_FPImage;k,l:longint);                 {  fi(y,x,l)*=fh(y   ,x   ,k), cut, undefined pixels assumed 1 }
begin
    SImg_Mul(cc,fh,fi,k,l,0,0);
end;


{*************************************************************************}
{******* SImg_Gauss ******************************************************}
{*************************************************************************}
procedure SImg_Gauss(cc:CSImg_Config; fi:TSImg_FPImage; k,xc,yc,rx,ry:longint; sx,sy:double);  { k:color; xc,yc: center; 2*rx+1 x 2*ry+1: size; sx,sy: sigma }
var i,j,m  : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Gauss: invalid image: fi');
    for i:=0 to fi.size-1 do
        fi.pt[i]:=0;
    sx:=1/(2*sqr(sx));
    sy:=1/(2*sqr(sy));
    while xc-rx<0 do
        xc+=fi.nx;
    while yc-ry<0 do
        yc+=fi.ny;
    for i:=-ry to ry do
    begin
        m:=((i+yc) mod fi.ny)*fi.sl+k;
        for j:=-rx to rx do
            fi.pt[m+((j+xc) mod fi.nx)*fi.nc]:=exp( -sy*i*i -sx*j*j );
    end;
end;


{************************************************************************}
{******* SImg_GetSum ****************************************************}
{************************************************************************}
function SImg_GetSum(cc:CSImg_Config; fi:TSImg_FPImage):double;
var i : longint;
    d : ^simgfp;
    r : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSum: illegal image: fi');
    r:=0;
    d:=fi.pt;
    for i:=0 to fi.size-1 do
        r+=d[i];
    result:=r;
end;

function SImg_GetSum(cc:CSImg_Config; fi:TSImg_FPImage; c:longint):double;
var i : longint;
    d : ^simgfp;
    r : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSum: illegal image: fi');
    if (c<0) or (c>=fi.nc) then cc.error('SImg_GetSumr: Invalid color: '+int2str(c));
    r:=0;
    d:=fi.pt+c;
    for i:=0 to fi.nx*fi.ny-1 do
        r+=d[i*fi.nc];
    result:=r;
end;


{*************************************************************************}
{******* SImg_GetMean ****************************************************}
{*************************************************************************}
function SImg_GetMean(cc:CSImg_Config; fi:TSImg_FPImage):double;
begin
    result:=SImg_GetSum(cc,fi)/fi.size;
end;

function SImg_GetMean(cc:CSImg_Config; fi:TSImg_FPImage; c:longint):double;
begin
    result:=SImg_GetSum(cc,fi,c)/(fi.nx*fi.ny);
end;


{*************************************************************************}
{******* SImg_AddScalar **************************************************}
{*************************************************************************}
procedure SImg_AddScalar(cc:CSImg_Config; fi:TSImg_FPImage; a:double);
var i : longint;
    d : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) or (fi.nc<>2) then cc.error('SImg_AddScalar: illegal image: fi');
d:=fi.pt;
for i:=0 to fi.size-1 do
 begin
 d^+=a; d+=1;
 end;
end;

procedure SImg_AddScalar(cc:CSImg_Config; fi:TSImg_FPImage; c:longint; a:double);
var i : longint;
    d : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_AddScalar: illegal image: fi');
    if (c<0) or (c>=fi.nc) then cc.error('SImg_AddScalar: Invalid color: '+int2str(c));
    d:=fi.pt+c;
    for i:=0 to fi.nx*fi.ny-1 do
        d[i*fi.nc]+=a;
end;


{*************************************************************************}
{******* SImg_MulScalar **************************************************}
{*************************************************************************}
procedure SImg_MulScalar(cc:CSImg_Config; fi:TSImg_FPImage; a:double);
var i : longint;
    d : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_MulScalar: illegal image: fi');
    d:=fi.pt;
    for i:=0 to fi.size-1 do
        d[i]*=a; d+=1;
end;

procedure SImg_MulScalar(cc:CSImg_Config; fi:TSImg_FPImage; c:longint; a:double);
var i : longint;
    d : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_MulScalar: illegal image: fi');
    if (c<0) or (c>=fi.nc) then cc.error('SImg_MulScalar: Invalid color: '+int2str(c));
    d:=fi.pt+c;
    for i:=0 to fi.nx*fi.ny-1 do
        d[i*fi.nc]*=a;
end;


{************************************************************************}
{******* SImg_CropNegaive ***********************************************}
{************************************************************************}
procedure SImg_CropNegative(cc:CSImg_Config; fi:TSImg_FPImage);
var i : longint;
    d : ^simgfp;
begin
    d:=fi.pt;
    for i:=0 to fi.size-1 do
    begin
        if d^<0 then d^:=0;
        d+=1;
    end;
end;

procedure SImg_CropNegative(cc:CSImg_Config; fi:TSImg_FPImage; c:longint);
var i : longint;
    d : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_CropNegative: illegal image: fi');
    if (c<0) or (c>=fi.nc) then cc.error('SImg_CropNegative: Invalid color: '+int2str(c));
    d:=fi.pt+c;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        if d^<0 then d^:=0;
        d+=fi.nc;
    end;
end;


{************************************************************************}
{******* SImg_Duplicate2 ************************************************}
{************************************************************************}
procedure SImg_Duplicate2(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; nc:longint);
var i,j   : longint;
    d,e   : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Duplicate12: illegal image: fi');
    if (fi.nc<>1) and (fi.nc<>nc) then cc.error('SImg_Duplicate2: Invalid number of colors: fi');
    if fi.nc=nc
    then SImg_Duplicate(cc,fi,fj,true)
    else begin
        SImg_New(cc,fj,fi.nx,fi.ny,nc);
        d:=fi.pt; e:=fj.pt;
        for i:=0 to fi.size-1 do
        begin
            for j:=0 to nc-1 do
                e[j]:=d^;
            e+=nc;
            d+=1;
        end;
    end;
end;
