{  simg -- visualization library
   Version 20210727
   Copyright (C) 1999-2021 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{ ************************************************************************* }
{ ******** siv_init ******************************************************* }
{ ************************************************************************* }
procedure siv_init(cc:CSImg_Config);
var s : simgstring;
begin
    gtk_rc_add_default_file(conf_sharedir+'/simgrc');
    s:=SImg_HomeDir+'/simgrc';
    gtk_rc_add_default_file(str2pchar(s));
    gtk_init(@argc, @argv);
    gdk_rgb_init;
    siv_prevConfig:=cc;
    siv_config:=CSImg_Config.create(cc);
    siv_config.SetMessageProc(@siv_print_message);
end;

{ ************************************************************************* }
{ ******** siv_main ******************************************************* }
{ ************************************************************************* }
procedure siv_main;
begin
    gtk_main;
end;

{ ************************************************************************* }
{ ******** siv_done ******************************************************* }
{ ************************************************************************* }
procedure siv_done;
begin
    siv_config.destroy;
end;

{ ************************************************************************* }
{ ******** siv_error ****************************************************** }
{ ************************************************************************* }
procedure siv_error(mes:simgstring);
begin
    siv_config.error(mes);
end;

{ ************************************************************************* }
{ ******** siv_warning **************************************************** }
{ ************************************************************************* }
procedure siv_warning(mes:simgstring);
begin
    siv_config.warning(mes);
end;

{ ************************************************************************* }
{ ******** siv_debug_begin ************************************************ }
{ ************************************************************************* }
procedure siv_debug_begin(mes:simgstring);
begin
    SImg_IncDebugLevel;;
    siv_config.debug(mes);
end;

procedure siv_debug_begin;
begin
    SImg_IncDebugLevel;;
end;

{ ************************************************************************* }
{ ******** siv_debug_end ************************************************** }
{ ************************************************************************* }
procedure siv_debug_end(mes:simgstring);
begin
    siv_config.debug(mes);
    SImg_DecDebugLevel;;
end;

procedure siv_debug_end;
begin
    SImg_DecDebugLevel;;
end;

{ ************************************************************************* }
{ ******** siv_OutputCatchedException ************************************* }
{ ************************************************************************* }
procedure SIV_OutputCatchedException;
begin
    SImg_OutputCatchedException(siv_config);
end;

{ ************************************************************************* }
{ ******** siv_set_rgb_color ********************************************** }
{ ************************************************************************* }
procedure siv_set_rgb_color(var c:TGdkColor; r,g,b:double);
begin
    c.pixel:=gdk_rgb_xpixel_from_rgb(round(r*255) shl 16 + round(g*255) shl 8 + round(b*255));
    c.red:=round(r*65535);
    c.green:=round(g*65535);
    c.blue:=round(b*65535);
end;

{ ************************************************************************* }
{ ******** siv_real2str *************************************************** }
{ ************************************************************************* }
function siv_real2str(r:double; i:longint):simgstring;
begin
    str(r:0:i,result);
    while result[length(result)]='0' do
        delete(result,length(result),1);
    if result[length(result)]='.' then delete(result,length(result),1);
end;

function siv_real2str(r:double; i,j:longint):simgstring;
begin
    str(r:0:i,result);
    while (result[length(result)]='0') or (length(result)>j) do
        delete(result,length(result),1);
    if result[length(result)]='.' then delete(result,length(result),1);
    while length(result)<j do
        result+=' ';
end;

{ ************************************************************************* }
{ ******** siv_label_new ************************************************** }
{ ************************************************************************* }
function siv_label_new(text:simgstring):PGtkWidget;
begin
    result:=gtk_label_new(str2pchar(text));
end;

{ ************************************************************************* }
{ ******** siv_button_new_with_label ************************************** }
{ ************************************************************************* }
function siv_button_new_with_label(text:simgstring):PGtkWidget;
begin
    result:=gtk_button_new_with_label(str2pchar(text));
end;

{ ************************************************************************* }
{ ******** siv_toggle_button_new_with_label ******************************* }
{ ************************************************************************* }
function siv_toggle_button_new_with_label(text:simgstring):PGtkWidget;
begin
    result:=gtk_toggle_button_new_with_label(str2pchar(text));
end;

{ ************************************************************************* }
{ ******** CSIV_History *************************************************** }
{ ************************************************************************* }
{ ******** init *********************************************************** }
constructor CSiv_History.init;
var i : longint;
begin
    for i:=0 to siv_history_size-1 do
        entries[i]:='';
end;

constructor CSiv_History.init(fn:simgstring);
var i,j : longint;
    s   : simgstring;
    t   : text;
begin
    j:=0;
    if FileExists(fn) then
    begin
        try
            assign(t,fn);
            reset(t);
            while (not eof(t)) and (j<siv_history_size) do
            begin
                readln(t,s);
                if s<>'' then
                    for i:=0 to j-1 do
                        if entries[i]=s then s:='';
                if s<>'' then
                begin
                    entries[j]:=s;
                    j+=1;
                end;
            end;
            close(t);
        except
            SIV_OutputCatchedException;
        end;
    end;
    for i:=j to siv_history_size-1 do
        entries[i]:='';
end;

constructor CSiv_History.init(h:CSIV_History);
var i : longint;
begin
    for i:=0 to siv_history_size-1 do
        entries[i]:=h.entries[i];
end;

{ ******** write ********************************************************** }
procedure CSIV_History.write(fn:simgstring);
var i  : longint;
    t  : text;
begin
    try
        assign(t,fn);
        rewrite(t);
        for i:=0 to siv_history_size-1 do
            if entries[i]<>'' then writeln(t,entries[i]);
        close(t);
    except
        SIV_OutputCatchedException;
    end;
end;

{ ******** insert ********************************************************* }
procedure CSiv_History.insert(s:simgstring);
var i : longint;
begin
    if s='' then exit;
    for i:=siv_history_size-1 downto 0 do
        if entries[i]=s then remove(i);
    for i:=siv_history_size-1 downto 1 do
        entries[i]:=entries[i-1];
    entries[0]:=s;
end;

{ ******** remove ********************************************************* }
procedure CSiv_History.remove(i:longint);
var j : longint;
begin
    if (i<0) or (i>=siv_history_size) then exit;
    for j:=i+1 to siv_history_size-1 do
        entries[j-1]:=entries[j];
    entries[siv_history_size-1]:='';
end;

{ ******** search ********************************************************* }
function CSiv_History.search(s:simgstring):longint;
var i,j : longint;
begin
    j:=-1;
    for i:=siv_history_size-1 downto 0 do
        if entries[i]=s then j:=i;
    result:=j;
end;

{ ******** count ********************************************************** }
function CSiv_History.count:longint;
var i,j : longint;
begin
    j:=0;
    for i:=0 to siv_history_size-1 do
        if entries[i]<>'' then j+=1;
    for i:=0 to j-1 do
        while entries[i]='' do
            remove(i);
    result:=j;
end;

{ ************************************************************************* }
{ ******** SIV_InitSys **************************************************** }
{ ************************************************************************* }
procedure SIV_InitSys;
begin
    siv_global_history:=CSiv_History.init(SImg_HomeDir+'/'+siv_history_filename);
end;

{ ************************************************************************* }
{ ******** SIV_DoneSys **************************************************** }
{ ************************************************************************* }
procedure SIV_DoneSys;
begin
    siv_global_history.write(SImg_HomeDir+'/'+siv_history_filename);
    siv_global_history.destroy;
end;

{ ************************************************************************* }
{ ******** CSIV_HScale **************************************************** }
{ ************************************************************************* }
{ ******** siv_hscale_stepcheck_idle ************************************** }
function siv_hscale_stepcheck_idle(self:CSIV_HScale):longint; cdecl;
begin
    result:=0;
    if self=nil then
    begin
        siv_warning('siv_hscale_stepcheck_idle: self=nil');
        exit;
    end;
    with self do
    begin
        if state_up and state_down then exit;
        if state_up then adj^.value+=adj^.step_increment;
        if state_down then adj^.value-=adj^.step_increment;
        if state_up or state_down then
        begin
            if adj^.value>adj^.upper then
            begin
                adj^.value:=adj^.upper;
                state_up:=false;
            end;
            if adj^.value<adj^.lower then
            begin
                 adj^.value:=adj^.lower;
                 state_down:=false;
            end;
            gtk_adjustment_value_changed(adj);
        end;
    end;
end;

{ ******** siv_hscale_enter_notify **************************************** }
function siv_hscale_enter_notify(ewidget:PGtkWidget; event:PGdkEventCrossing; self:CSIV_HScale):longint; cdecl;
begin
    result:=0;
    if self=nil then
    begin
        siv_warning('siv_hscale_enter_notify: self=nil');
        exit;
    end;
    with self do
    begin
        if ewidget=step_up then gtk_image_set_from_pixbuf(PGtkImage(arrow_up),arrow_up_xpm[1])
        else if ewidget=step_down then gtk_image_set_from_pixbuf(PGtkImage(arrow_down),arrow_down_xpm[1])
        else siv_warning('siv_hscale_enter_notify: Unknown widget');
    end;
    result:=1;
end;

{ ******** siv_hscale_leave_notify **************************************** }
function siv_hscale_leave_notify(ewidget:PGtkWidget; event:PGdkEventCrossing; self:CSIV_HScale):longint; cdecl;
begin
    result:=0;
    if self=nil then
    begin
        siv_warning('siv_hscale_leave_notify: self=nil');
        exit;
    end;
    with self do
    begin
        if ewidget=step_up then
        begin
            gtk_image_set_from_pixbuf(PGtkImage(arrow_up),arrow_up_xpm[0]);
            state_up:=false;
        end
        else if ewidget=step_down then
        begin
            gtk_image_set_from_pixbuf(PGtkImage(arrow_down),arrow_down_xpm[0]);
            state_down:=false;
        end
        else siv_warning('siv_hscale_leave_notify: Unknown widget');
    end;
    result:=1;
end;

{ ******** siv_hscale_button_press **************************************** }
function siv_hscale_button_press(ewidget:PGtkWidget; event:PGdkEventButton; self:CSIV_HScale):longint; cdecl;
begin
    result:=0;
    if self=nil then
    begin
        siv_warning('siv_hscale_button_press: self=nil');
        exit;
    end;
    with self do
    begin
        if ewidget=step_up then
        begin
            state_up:=true;
            steptime:=500;
            siv_hscale_stepcheck_idle(self);
        end
        else if ewidget=step_down then
        begin
            state_down:=true;
            steptime:=500;
            siv_hscale_stepcheck_idle(self);
        end
        else siv_warning('siv_hscale_button_press: Unknown widget');
    end;
    result:=1;
end;

{ ******** siv_hscale_button_release ************************************** }
function siv_hscale_button_release(ewidget:PGtkWidget; event:PGdkEventButton; self:CSIV_HScale):longint; cdecl;
begin
    result:=0;
    if self=nil then
    begin
        siv_warning('siv_hscale_button_release: self=nil');
        exit;
    end;
    with self do
    begin
        if ewidget=step_up then
        begin
//            if state_up then gtk_widget_set_state(arrow_up,GTK_STATE_PRELIGHT)
//            else gtk_widget_set_state(arrow_up,GTK_STATE_NORMAL);
            state_up:=false;
        end
        else if ewidget=step_down then
        begin
//            if state_down then gtk_widget_set_state(arrow_down,GTK_STATE_PRELIGHT)
//            else gtk_widget_set_state(arrow_down,GTK_STATE_NORMAL);
            state_down:=false;
        end
        else siv_warning('siv_hscale_button_release: Unknown widget');
    end;
    result:=1;
end;


{ ******** methodes ******************************************************* }
{ ******** init *********************************************************** }
constructor CSIV_HScale.init(parent:PGtkWidget; value,lower,upper,step_increment,page_increment:double);
begin
    { **** hbox **** }
    hbox:=gtk_hbox_new(false,0);
    { **** adj ****}
    adj:=PGtkAdjustment(gtk_adjustment_new(value,lower,upper,step_increment,page_increment,0));
    { **** hscale **** }
    hscale:=gtk_hscale_new(adj);
    gtk_scale_set_digits(GTK_SCALE(hscale), 3);
    gtk_scale_set_draw_value(PGtkScale(hscale),false);
    gtk_range_set_update_policy(PGtkRange(hscale),GTK_UPDATE_DISCONTINUOUS);
    gtk_box_pack_start(GTK_BOX(hbox), hscale, true, true, 0);
    gtk_widget_show(hscale);

    { **** vbox **** }
    vbox:=gtk_vbox_new(false,0);
    gtk_box_pack_start(GTK_BOX(hbox), vbox, false,false, 0);
    gtk_widget_show(vbox);

    { ** step_up ** }
    arrow_up_xpm[0]:=gdk_pixbuf_new_from_xpm_data(@xpm_up_arrow1);
    arrow_up_xpm[1]:=gdk_pixbuf_new_from_xpm_data(@xpm_up_arrow2);
    arrow_up:=gtk_image_new_from_pixbuf(arrow_up_xpm[0]);
    gtk_widget_show(arrow_up);
    step_up:=gtk_event_box_new;
    gtk_container_add(PGtkContainer(step_up),arrow_up);
    gtk_widget_show(step_up);
    gtk_box_pack_start(GTK_BOX(vbox), step_up, false, false, 0);
    state_up:=false;
    gtk_widget_set_flags(step_up,GTK_CAN_FOCUS);
    gtk_widget_set_events(step_up,GDK_BUTTON_PRESS_MASK
        or GDK_ENTER_NOTIFY
        or GDK_LEAVE_NOTIFY
        or GDK_BUTTON_RELEASE_MASK);
    g_signal_connect(GTK_OBJECT(step_up),'enter-notify-event',TGtkSignalFunc(@siv_hscale_enter_notify),self);
        g_signal_connect(GTK_OBJECT(step_up),'leave-notify-event',TGtkSignalFunc(@siv_hscale_leave_notify),self);
    g_signal_connect(GTK_OBJECT(step_up),'button-press-event',TGtkSignalFunc(@siv_hscale_button_press),self);
    g_signal_connect(GTK_OBJECT(step_up),'button-release-event',TGtkSignalFunc(@siv_hscale_button_release),self);

    { ** step_down ** }
    arrow_down_xpm[0]:=gdk_pixbuf_new_from_xpm_data(@xpm_down_arrow1);
    arrow_down_xpm[1]:=gdk_pixbuf_new_from_xpm_data(@xpm_down_arrow2);
    arrow_down:=gtk_image_new_from_pixbuf(arrow_down_xpm[0]);
    gtk_widget_show(arrow_down);
    step_down:=gtk_event_box_new;
    gtk_container_add(PGtkContainer(step_down),arrow_down);
    gtk_widget_show(step_down);
    gtk_box_pack_start(GTK_BOX(vbox), step_down, false, false, 0);
    state_down:=false;
    gtk_widget_set_flags(step_down,GTK_CAN_FOCUS);
    gtk_widget_set_events(step_down,GDK_BUTTON_PRESS_MASK
        or GDK_ENTER_NOTIFY
        or GDK_LEAVE_NOTIFY
        or GDK_BUTTON_RELEASE_MASK);
    g_signal_connect(GTK_OBJECT(step_down),'enter-notify-event',TGtkSignalFunc(@siv_hscale_enter_notify),self);
    g_signal_connect(GTK_OBJECT(step_down),'leave-notify-event',TGtkSignalFunc(@siv_hscale_leave_notify),self);
    g_signal_connect(GTK_OBJECT(step_down),'button-press-event',TGtkSignalFunc(@siv_hscale_button_press),self);
    g_signal_connect(GTK_OBJECT(step_down),'button-release-event',TGtkSignalFunc(@siv_hscale_button_release),self);

    thelabel:=nil;
end;

{ ******** done *********************************************************** }
destructor CSIV_HScale.done;
begin
end;

{ ******** widget ********************************************************* }
function CSIV_HScale.widget:PGtkWidget;
begin
    result:=hbox;
end;

{ ******** set_label ****************************************************** }
procedure CSIV_HScale.set_label(xsize:longint; s:simgstring);
var align  : PGtkWidget;
begin
    s:=' '+s;
    if thelabel=nil then
    begin
        thelabel:=gtk_label_new(str2pchar(s));
        gtk_widget_show(thelabel);
        if xsize>8 then
        begin
            align:=gtk_alignment_new(0,0,0,0);
            gtk_widget_set_usize(align,xsize,8);
            gtk_widget_show(align);
            gtk_box_pack_start(GTK_BOX(hbox),align,false,false,0);
            thelabel:=gtk_label_new(str2pchar(s));
            gtk_widget_show(thelabel);
            gtk_container_add(PGtkContainer(align),thelabel);
        end
        else begin
            thelabel:=gtk_label_new(str2pchar(s));
            gtk_widget_show(thelabel);
            gtk_box_pack_start(GTK_BOX(hbox),thelabel,false,false,0);
        end;
    end
    else gtk_label_set_text(PGtkLabel(thelabel),str2pchar(s));
end;

{ ******** set_tooltip **************************************************** }
procedure CSIV_HScale.set_tooltip(s:simgstring);
begin
    siv_tooltip_new(hscale,str2pchar(s),nil);
end;

{ ******** connect ******************************************************** }
procedure CSIV_HScale.connect(proc:TGtkSignalFunc; data:gpointer);
begin
    g_signal_connect(GTK_OBJECT(adj), 'value_changed', GTK_SIGNAL_FUNC(proc), data);
end;

{ ******** stepcheck ****************************************************** }
procedure CSIV_HScale.stepcheck;
begin
    if state_up or state_down then gtk_timeout_add(steptime,TGtkFunction(@siv_hscale_stepcheck_idle),self);
    steptime:=(steptime+25) shr 1;
end;

{ ******** show *********************************************************** }
procedure CSIV_HScale.show;
begin
    gtk_widget_show(hbox);
end;

{ ******** hide *********************************************************** }
procedure CSIV_HScale.hide;
begin
    gtk_widget_hide(hbox);
end;

